import torch
from torch.nn import functional as F
import time
from sklearn.metrics import cohen_kappa_score
import plotly.graph_objects as go
from functools import lru_cache
import pandas as pd
import numpy as np

if torch.cuda.is_available():
    from pynvml.smi import nvidia_smi
    smi_instance = nvidia_smi.getInstance()


@lru_cache(8)
def logit_to_label(logits, threshold=None):
    """Convert logits into predicted labels. This function uses a least-recently-used cache to
    avoid recalculating predictions for multiple metric functions.

    Arguments:
        logits {torch.Tensor} -- Tensor of logits of size (batch_size x num_classes).

    Keyword Arguments:
        threshold {float} -- Optional threshold value to use in converting to labels. Only valid
            when num_classes = 2. (default: {None})

    Returns:
        torch.Tensor -- Tensor of predicted labels of length batch_size.
    """
    if threshold is not None:
        assert logits.shape[1] == 2, "Probability threshold only valid for binary classification"
        probs = F.softmax(logits, dim=1)
        preds = (probs[:, 1] >= float(threshold)).long()
    else:
        preds = logits.argmax(dim=1)
    return preds


def _kappa(logits, y, threshold=0.5):
    preds = logit_to_label(logits, threshold)
    return torch.as_tensor(cohen_kappa_score(preds.detach().cpu(), y.detach().cpu()))


def _accuracy(logits, y, threshold=0.5):
    preds = logit_to_label(logits, threshold)
    return (preds == y).float().mean()


def _auc(fpr, tpr):
    """Calculate AUC given FPR and TPR values.
    
    Note that this function assumes the FPR and TPR values are sorted according to monotonically 
    increasing probability thresholds.
    """
    widths = fpr[:-1] - fpr[1:]
    heights = (tpr[:-1] + tpr[1:]) / 2
    return (widths * heights).sum()


@lru_cache(32)
def _crosstab(preds, y, binary=False):
    dev = 'cpu' if preds.get_device() == -1 else preds.get_device()

    if binary:
        correct = preds == y
        y = y.bool()
        cm = torch.zeros(2, 2, device=dev)
        cm[0, 0] = (correct & ~y).sum()
        cm[0, 1] = (~correct & ~y).sum()
        cm[1, 0] = (~correct & y).sum()
        cm[1, 1] = (correct & y).sum()

    else:
        y_unique = torch.unique(torch.cat((preds, y)).int()).sort()[0]
        cm = torch.zeros(y_unique[-1] + 1, y_unique[-1] + 1, device=dev)
        for cls_y in y_unique:
            for cls_yp in y_unique:
                cm[cls_y, cls_yp] = ((y == cls_y) & (preds == cls_yp)).sum()

    return cm


def _confusion_matrix(logits, y, threshold=None):
    preds = logit_to_label(logits, threshold=threshold)
    return _crosstab(preds, y)


def _confusion_matrix_array(logits, y, thresholds, do_softmax=True):
    """For binary classification only - an intermediate step for ROC calculation."""
    dev = 'cpu' if y.get_device() == -1 else y.get_device()
    thresholds = torch.as_tensor(thresholds).to(dev)

    # Get probabilities
    if do_softmax:
        probs = F.softmax(logits, dim=1)[:, 1]
    else:
        probs = logits[:, 1]

    # For efficiency, find all thresholds at which the CM will actually change
    thresh_incr = thresholds[1] - thresholds[0]
    probs_trunc = (probs / thresh_incr).trunc() * thresh_incr
    steps = (thresholds.unsqueeze(1) - probs_trunc.unsqueeze(0)).abs().argmin(dim=0) + 1
    steps = [0] + steps.unique().tolist()

    cm_array = torch.zeros(len(thresholds), 2, 2, device=dev)
    for i, threshold in enumerate(thresholds):
        if i in steps:
            preds = (probs >= float(threshold)).long()
            cm = _crosstab(preds, y, binary=True)
        cm_array[i] = cm

    return cm_array


def _generate_plot(x, y, text, xlabel, ylabel, label, fig, transparent=False):
    if fig is None:
        fig = go.Figure()
        fig.update_layout(
            xaxis_title=xlabel,
            yaxis_title=ylabel,
            xaxis={'range': [-0.01, 1.005]},
            yaxis={'scaleanchor': "x", 'scaleratio': 1, 'range': [-0.01, 1.0]},
            height=600, width=690,
            margin={'t': 10, 'b': 10, 'l': 10, 'r': 10},
        )
        if transparent:
            fig.update_layout(
                paper_bgcolor='rgba(0,0,0,0)',
                plot_bgcolor='rgba(0,0,0,0)',
                font={'color': '#808080'},
                xaxis={'gridcolor': '#CCCCCC', 'zerolinecolor': '#CCCCCC'},
                yaxis={'gridcolor': '#CCCCCC', 'zerolinecolor': '#CCCCCC'},
            )

    fig.add_trace(go.Scatter(x=x, y=y, text=text, name=label))
    
    return fig


class Metric(object):
    """Base class for creating metrics that use a cache of passed values (e.g., confusion matrix
    cells).

    Raises:
        NotImplementedError: when __call__ not defined.
        NotImplementedError: when reset not defined.
    """

    def __init__(self):
        """Initialize cached values."""
        self.reset()

    def __call__(self):
        """Update cache and return current (batch-specific) metric value."""
        raise NotImplementedError

    def compute(self):
        """Return overall metric value, combined across all batches."""
        raise NotImplementedError

    def reset(self):
        """Reset cached values."""
        raise NotImplementedError


class PooledMean(Metric):
    """Base class for converting a standard callable into a pooled mean metric.

    This class will store results from calls so that the overall mean can be calculated in 
    MapReduce fashion.

    E.g.,
    First __call__: fn(y_pred, y) -> 10.4, len(y) -> 32
    Second __call__: fn(y_pred, y) -> 3.4, len(y) -> 24
    compute: (10.4 * 32 + 3.4 * 24) / (32 + 24)

    Arguments:
        fn {callable} -- A function that returns a torch scalar metric value. It should have the 
            signature `fn(y_pred: torch.Tensor, y: torch.Tensor) -> torch.Tensor`.
    """

    def __init__(self, fn):
        self.fn = fn
        super().__init__()

    def __call__(self, y_pred, y):
        self.latest_value = self.fn(y_pred, y)
        self.latest_num_samples = float(len(y_pred))
        self.num_samples += self.latest_num_samples
        self.value_sum += self.latest_value.detach() * self.latest_num_samples
        return self.latest_value

    def compute(self):
        return self.value_sum / self.num_samples

    def reset(self):
        self.value_sum = 0
        self.num_samples = 0
        self.latest_value = 0
        self.latest_num_samples = 0


class EPS(Metric):
    """Examples per second.

    Use this class for tracking training and testing samples per second. Note that the __call__
    method expects two arguments (the predictions and actuals), so that its signature matches the
    loss and metric functions. This means that it can be used in the same way. In addition, the
    timing will be returned as a torch tensor object.

    Example:
    >>> timer = EPS()
    >>> for x, y in loader:
    >>>     y_bar = model(x)
    >>>     loss = loss_fn(y_bar, y)
    >>>     print('EPS: ', timer(y_bar, y))
    """

    def __call__(self, y_pred, y):
        end = time.time()
        self.latest_elapsed = end - self.start
        self.latest_num_samples = float(len(y_pred))
        self.elapsed += self.latest_elapsed
        self.num_samples += self.latest_num_samples
        self.start = end

        return torch.tensor(self.latest_num_samples / self.latest_elapsed)

    def compute(self):
        return torch.tensor(self.num_samples / self.elapsed)

    def reset(self):
        self.start = time.time()
        self.elapsed = 0
        self.num_samples = 0
        self.latest_elapsed = 0
        self.latest_num_samples = 0


class Accuracy(PooledMean):
    """Accuracy metric."""

    def __init__(self, threshold=None):
        if threshold is not None and abs(threshold - 0.5) < 1e-5:
            threshold = None
        fn = lambda y_pred, y: _accuracy(y_pred, y, threshold)
        super().__init__(fn)


class Kappa(PooledMean):
    """Cohen's Kappa metric."""

    def __init__(self):
        super().__init__(_kappa)


class ConfusionMatrix(Metric):
    """Confusion matrix. During training with a runner, this metric will not be logged, but the
    overall confusion matrix can be accessed from the Confusion matrix object later (it is stored
    in the `value` attribute).
    
    Use ConfusionMatrix.print() to print the confusion matrix.
    """

    def __init__(self, threshold=None, classnames=None):
        self.classnames = classnames
        self.fn = lambda *x: _confusion_matrix(*x, threshold=threshold)
        super().__init__()

    def __call__(self, y_pred, y):
        self.value += self.fn(y_pred, y)

    def compute(self):
        pass

    def reset(self):
        self.value = 0
    
    def print(self):
        mat = self.value.detach().cpu().numpy()
        if self.classnames is None:
            classnames = [str(i) for i in range(len(mat))]
        else:
            classnames = self.classnames
        print(pd.DataFrame(mat, columns=classnames, index=classnames))


class _ConfusionMatrixCurve(Metric):
    """Base class for metrics that utilize confusion matrices (e.g., ROC AUC).
    
    Keyword Arguments:
        increment {float} -- Probability increment. (default: {0.02})
    """

    def __init__(self, increment=0.02):
        self.probs = torch.arange(0, 1+1e-8, increment)
        self.fn = lambda y_pred, y: _confusion_matrix_array(y_pred, y, self.probs)
        super().__init__()

    def __call__(self, y_pred, y):
        self.latest_value = self.fn(y_pred, y)
        self.latest_num_samples = float(len(y_pred))
        self.num_samples += self.latest_num_samples
        self.value_sum += self.latest_value.detach()
        return self._compute_values(self.latest_value.detach())
    
    def compute(self):
        return self._compute_values(self.value_sum)

    def reset(self):
        self.value_sum = 0
        self.num_samples = 0
        self.latest_value = 0
        self.latest_num_samples = 0
    
    def _compute_values(self, cms):
        raise NotImplementedError


class ROC_AUC(_ConfusionMatrixCurve):
    """Metric class to iteratively calculate the ROC curve and AUC.
    
    Keyword Arguments:
        increment {float} -- Probability increment. (default: {0.02})
    """

    def __call__(self, y_pred, y):
        self.latest_value = self.fn(y_pred, y)
        self.latest_num_samples = float(len(y_pred))
        self.num_samples += self.latest_num_samples
        self.value_sum += self.latest_value.detach()
        return torch.as_tensor(self._compute_values(self.latest_value.detach())[0])
    
    def compute(self):
        return torch.as_tensor(self._compute_values(self.value_sum)[0])
    
    def plot(self, fig=None, curve_name=''):
        auc_score, fpr, tpr = self._compute_values(self.value_sum)

        fig = _generate_plot(
            fpr.cpu(), tpr.cpu(), np.array(self.probs).round(3).astype(str),
            'False positive rate', 'True positive rate',
            f"{curve_name} (AUC: {auc_score:.3f})", fig
        )

        return fig
    
    def _compute_values(self, cms):
        negatives = cms[0, 0, 1]
        positives = cms[0, 1, 1]

        fpr = cms[:, 0, 1] / (negatives + 1e-6)
        tpr = cms[:, 1, 1] / (positives + 1e-6)

        auc_score = _auc(fpr, tpr)

        return auc_score, fpr, tpr


class BestAccuracy(_ConfusionMatrixCurve):
    """Metric class to iteratively accumulate correct counts for various thresholds and return the
    best possible accuracy.
    
    Keyword Arguments:
        increment {float} -- Probability increment. (default: {0.02})
    """
    
    def _compute_values(self, cms):
        correct = cms[:, 0, 0] + cms[:, 1, 1]
        count = cms[0].sum().float()

        accuracy = correct.float() / count
        accuracy, best_ind = torch.max(accuracy, dim=0)

        self.best_prob = self.probs[best_ind]

        return accuracy


class FPR(_ConfusionMatrixCurve):
    """Metric class to iteratively calculate false positive rate for a given true positive rate.
    
    Keyword Arguments:
        tpr {float} -- Reference true positive rate value. (default: {0.9})
        increment {float} -- Probability increment. (default: {0.02})
    """
    
    def __init__(self, tpr=0.9, increment=0.02):
        self.tpr = tpr
        super().__init__(increment)

    def _compute_values(self, cms):
        negatives = cms[0, 0, 1]
        positives = cms[0, 1, 1]

        if negatives > 1 and positives > 1:
            fpr = cms[:, 0, 1] / negatives
            tpr = cms[:, 1, 1] / positives
        
            ind = torch.nonzero((tpr - self.tpr) >= 1e-6, as_tuple=True)[0][-1]
            
            return fpr[ind]
        else:
            return torch.tensor(float('nan'))


class TPR(_ConfusionMatrixCurve):
    """Metric class to iteratively calculate true positive rate for a given false positive rate.
    
    Keyword Arguments:
        fpr {float} -- Reference false positive rate value. (default: {0.1})
        increment {float} -- Probability increment. (default: {0.02})
    """
    
    def __init__(self, fpr=0.1, increment=0.02):
        self.fpr = fpr
        super().__init__(increment)

    def _compute_values(self, cms):
        negatives = cms[0, 0, 1]
        positives = cms[0, 1, 1]

        if negatives > 1 and positives > 1:
            fpr = cms[:, 0, 1] / negatives
            tpr = cms[:, 1, 1] / positives

            ind = torch.nonzero((fpr - self.fpr) <= 1e-6, as_tuple=True)[0][0]

            return tpr[ind]
        else:
            return torch.tensor(float('nan'))


def get_gpu_util(*unused):
    if torch.cuda.is_available():
        out = smi_instance.DeviceQuery('utilization.gpu')['gpu']
        out = [v['utilization']['gpu_util'] for v in out]
        out = torch.as_tensor(out).float().mean()
    else:
        out = torch.as_tensor(0.)
    return out


def get_gpu_mem(*unused):
    if torch.cuda.is_available():
        out = smi_instance.DeviceQuery('memory.used, memory.total')['gpu']
        out = [v['fb_memory_usage']['used'] / v['fb_memory_usage']['total'] * 100 for v in out]
        out = torch.as_tensor(out).mean()
    else:
        out = torch.as_tensor(0.)
    return out


class GPUUtil(PooledMean):
    """GPU utilization metric."""

    def __init__(self):
        super().__init__(get_gpu_util)


class GPUMem(PooledMean):
    """GPU memory metric."""

    def __init__(self):
        super().__init__(get_gpu_mem)


# Aliases for backward compatibility
accuracy = Accuracy()
kappa = Kappa()
BatchTimer = EPS



import os
import shutil
from importlib.util import spec_from_file_location, module_from_spec

import torch
from xt_training import Runner, metrics
from xt_training.utils import _import_config, Tee


def default_exit(config, runner, save_dir):
    pass


def test(args):
    config_path = args.config_path
    checkpoint_path = args.checkpoint_path
    save_dir = args.save_dir

    if os.path.isdir(config_path):
        config_dir = config_path
        config_path = os.path.join(config_dir, 'config.py')
        assert checkpoint_path is None, (
            "checkpoint_path is not valid when config_path is a directory.\n"
            "\tSpecify either a config script and (optional) checkpoint individually, or\n"
            "\tspecify a directory containing both config.py and best.pt"
        )
        checkpoint_path = os.path.join(config_dir, 'best.pt')
        if not save_dir:
            save_dir = config_dir

    # Initialize logging
    if save_dir:
        os.makedirs(save_dir, exist_ok=True)
        try:
            shutil.copy(config_path, f'{save_dir}/config.py')
        except shutil.SameFileError:
            pass
        tee = Tee(os.path.join(save_dir, "test.log"))
        results = {}

    config = _import_config(config_path)

    #  Load definitions
    val_loader = getattr(config, 'val_loader', None)
    test_loaders = getattr(config, 'test_loaders', None)
    model = config.model
    loss_fn = getattr(config, 'loss_fn', lambda *_: torch.tensor(0.))
    eval_metrics = getattr(config, 'eval_metrics', {'eps': metrics.EPS()})
    on_exit = getattr(config, 'test_exit', default_exit)

    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    print('Running on device: {}'.format(device))
    model = model.to(device)
    
    if checkpoint_path is not None:
        model.load_state_dict(torch.load(checkpoint_path))
    model.eval()

    # Define model runner
    runner = Runner(model, loss_fn, batch_metrics=eval_metrics, device=device)

    if val_loader:
        print('\n\nValidation')
        print('-' * 10)
        if save_dir:
            preds, labels = runner(val_loader, 'valid', return_preds=True)
            results['valid'] = {'preds': preds, 'labels': labels}
        else:
            runner(val_loader, 'valid')

    if test_loaders:
        print('\nTest')
        print('-' * 10)
        for loader_name, loader in test_loaders.items():
            if save_dir:
                preds, labels = runner(loader, loader_name, return_preds=True)
                results[loader_name] = {'preds': preds, 'labels': labels}
            else:
                runner(loader, loader_name)

    on_exit(config, runner, save_dir)

    if save_dir:
        torch.save(results, os.path.join(save_dir, 'results.pt'))
        tee.flush()
        tee.close()



import os
import shutil
from importlib.util import spec_from_file_location, module_from_spec


def template(args):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    shutil.copy(os.path.join(dir_path, 'config_example.py'), args.save_dir)


def _import_config(path):
    """Import a config file given a file path."""
    config_spec = spec_from_file_location("config", path)
    config = module_from_spec(config_spec)
    config_spec.loader.exec_module(config)

    return config



import sys


class Tee(object):
    """Object that replicates the functionality of the `tee` shell command."""

    def __init__(self, logfile):
        """Constructor for Tee class.
        
        Arguments:
            logfile {str} -- Path to destination file.
        """
        self.file = open(logfile, "w")
        self.stdout = sys.stdout
        sys.stdout = self

    def close(self):
        if self.stdout is not None:
            sys.stdout = self.stdout
            self.stdout = None
        if self.file is not None:
            self.file.close()
            self.file = None

    def write(self, data):
        self.file.write(data)
        self.stdout.write(data)

    def flush(self):
        self.file.flush()
        self.stdout.flush()

    def __del__(self):
        self.close()


import argparse

from .utils import train, template, test


def parse_args():
    parser = argparse.ArgumentParser(
        description='A tool for training/evaluating pytorch deep learning models'
    )
    subparsers = parser.add_subparsers(help='Subcommand help')

    parser_train = subparsers.add_parser(
        'train',
        help='Train a model given a config script.',
        description='Train a model given a config script.'
    )
    parser_train.add_argument(
        'config_path',
        type=str,
        help='Path to fully specified config.'
    )
    parser_train.add_argument(
        'save_dir',
        type=str,
        help='Path to save model checkpoints, tensorboard runs, logs, and config files.'
    )
    parser_train.add_argument(
        '--overwrite', '-o',
        action='store_true',
        help='Switch to allow overwriting of existing path if necessary.'
    )
    parser_train.set_defaults(func=train)

    parser_test = subparsers.add_parser(
        'test',
        help='Test a model given a config script.',
        description='Test a model given a config script.'
    )
    parser_test.add_argument(
        'config_path',
        type=str,
        help='Path to fully specified config (if a file) or checkpoint directory (if a directory).'
    )
    parser_test.add_argument(
        'checkpoint_path',
        type=str,
        help='Path to compatible model checkpoint.',
        default=None,
        nargs='?'
    )
    parser_test.add_argument(
        '--save_dir',
        type=str,
        help='Path to save logs, config files, and (optionally) predictions.',
        default=None
    )
    parser_test.set_defaults(func=test)

    parser_template = subparsers.add_parser(
        'template',
        help='Generate a config script template.',
        description='Generate a config script template.'
    )
    parser_template.add_argument(
        'save_dir',
        type=str,
        help='Path to save config template. This should end in ".py"'
    )
    parser_template.set_defaults(func=template)
    
    args = parser.parse_args()

    return args


if __name__ == '__main__':
    args = parse_args()
    args.func(args)



import os
import shutil
from importlib.util import spec_from_file_location, module_from_spec

import torch
from torch.utils.tensorboard import SummaryWriter
from xt_training import Runner, metrics
from xt_training.utils import _import_config, Tee


def default_exit(config, runner, save_dir):
    test_loaders = getattr(config, 'test_loaders', None)

    # Final evaluation against test set(s)
    if test_loaders:
        print('\nTest')
        print('-' * 10)
        config.model.eval()
        for loader_name, loader in test_loaders.items():
            runner(loader, loader_name)


def train(args):
    config_path = args.config_path
    save_dir = args.save_dir
    overwrite = args.overwrite

    # Initialize logging
    if overwrite and os.path.isdir(save_dir):
        shutil.rmtree(save_dir)
    os.makedirs(save_dir, exist_ok=True)
    tee = Tee(os.path.join(save_dir, "train.log"))

    if isinstance(config_path, str):
        config = _import_config(config_path)
        shutil.copy(config_path, f'{save_dir}/config.py')
    else:
        config = config_path
        shutil.copy(config['__file__'], f'{save_dir}/config.py')

    #  Load definitions
    train_loader = config.train_loader
    val_loader = getattr(config, 'val_loader', None)
    test_loaders = getattr(config, 'test_loaders', None)
    model = config.model
    optimizer = config.optimizer
    epochs = config.epochs
    scheduler = getattr(config, 'scheduler', None)
    loss_fn = config.loss_fn
    eval_metrics = getattr(config, 'eval_metrics', {'eps': metrics.EPS()})
    on_exit = getattr(config, 'train_exit', default_exit)

    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    print('Running on device: {}'.format(device))
    model = model.to(device)

    # Create tensorboard writer
    writer = SummaryWriter(save_dir, flush_secs=30)

    # Define model runner
    runner = Runner(
        model, loss_fn, optimizer, scheduler, batch_metrics=eval_metrics,
        device=device, writer=writer
    )

    if test_loaders:
        print('\n\nInitial')
        print('-' * 10)
        model.eval()
        for loader_name, loader in test_loaders.items():
            runner(loader, loader_name)

    best_loss = 1e12
    if val_loader:
        model.eval()
        runner(val_loader, 'valid')
        best_loss = runner.loss()

    torch.save(model.state_dict(), f'{save_dir}/latest.pt')
    shutil.copy(f'{save_dir}/latest.pt', f'{save_dir}/best.pt')

    try:
        for epoch in range(epochs):
            print('\nEpoch {}/{}'.format(epoch + 1, config.epochs))
            print('-' * 10) 

            if hasattr(model, 'update') and callable(model.update):
                model.update(epoch + 1)

            model.train()
            runner(train_loader, 'train')

            if val_loader:
                model.eval()
                runner(val_loader, 'valid')

            torch.save(model.state_dict(), f'{save_dir}/latest.pt')
            if runner.loss() < best_loss:
                shutil.copy(f'{save_dir}/latest.pt', f'{save_dir}/best.pt')
                best_loss = runner.loss()
                print(f'Saved new best: {best_loss:.4}')

    # Allow safe interruption of training loop
    except KeyboardInterrupt:
        print('\n\nExiting with honour\n')
        pass

    except Exception as e:
        print('\n\nDishonourable exit\n')
        raise e

    on_exit(config, runner, save_dir)

    writer.close()
    tee.flush()
    tee.close()



"""Pytorch deep learning configuration script.

Usage:
>>> python -m xt_training train <path_to_this_file> ...
OR
>>> python -m xt_training test <path_to_this_file> ...

For `python -m xt_training train`:
  Required:
    * train_loader
    * model
    * loss_fn
    * optimizer
    * epochs
  Optional:
    * val_loader
    * test_loaders
    * eval_metrics
    * scheduler
    * train_exit
  Unused:
    * test_exit

For `python -m xt_training test`:
  Required:
    * model
  Optional:
    * loss_fn
    * val_loader
    * test_loaders
    * eval_metrics
    * test_exit
  Unused:
    * train_loader
    * optimizer
    * epochs
    * scheduler
    * train_exit
"""

import torch
from torch import optim
from torch import nn
from torch.utils.data import DataLoader, SequentialSampler, RandomSampler
from torch.optim import lr_scheduler
from torchvision import models, transforms
import numpy as np
from xt_training import metrics
from xt_training.utils import training, testing

# Dataset
train_dataset = torch.utils.data.TensorDataset(
    torch.rand(128, 3, 224, 224),
    torch.randint(1000, (128,))
)
val_dataset = torch.utils.data.TensorDataset(
    torch.rand(16, 3, 224, 224),
    torch.randint(1000, (16,))
)
test_datasets = {
    'test': torch.utils.data.TensorDataset(
        torch.rand(16, 3, 224, 224),
        torch.randint(1000, (16,))
    )
}

# Dataloaders
batch_size = 8
num_workers = 4
train_loader = DataLoader(
    train_dataset,
    batch_size=batch_size,
    num_workers=num_workers,
    sampler=RandomSampler(train_dataset)
)
val_loader = DataLoader(
    val_dataset,
    batch_size=batch_size,
    num_workers=num_workers,
    sampler=SequentialSampler(val_dataset)
)
test_loaders = {}
for ds_name, ds in test_datasets.items():
    test_loaders[ds_name] = DataLoader(
        ds,
        batch_size=batch_size,
        num_workers=num_workers,
    )

# Model
model = models.resnet18(pretrained=True)

# Loss and metrics
loss_fn = nn.CrossEntropyLoss()
eval_metrics = {
    'acc': metrics.Accuracy(),
    'util': metrics.GPUUtil(),
    'mem': metrics.GPUMem()
}

# Optimizer
optimizer = optim.Adam(model.parameters(), lr=1e-3)

# Scheduler
epochs = 10
scheduler = lr_scheduler.MultiStepLR(optimizer, milestones=[5])


# Function to run after training
def train_exit(config, runner, save_dir):
    training.default_exit(config, runner, save_dir)


# Function to run after testing
def test_exit(config, runner, save_dir):
    testing.default_exit(config, runner, save_dir)



import setuptools, os

PACKAGE_NAME = 'xt-training'
VERSION = '1.8.1'
AUTHOR = 'Xtract AI'
EMAIL = 'info@xtract.ai'
DESCRIPTION = 'Utilities for training models in pytorch'
GITHUB_URL = 'https://github.com/XtractTech/xt-training'

parent_dir = os.path.dirname(os.path.realpath(__file__))

with open(f'{parent_dir}/README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name=PACKAGE_NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=GITHUB_URL,
    packages=[
        'xt_training',
        'xt_training.utils'
    ],
    provides=['xt_training'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy',
        'pandas',
        'torch',
        'scikit-learn',
        'plotly',
        'pynvml',
    ],
)



import torch
import numpy as np
import json

from collections import Iterable
from .metrics import EPS, PooledMean, Metric

class Logger(object):

    def __init__(self, mode, length):
        """Text logging class.

        Arguments:
            mode {str} -- Run mode, used as a prefix in log output (e.g., 'train' or 'valid').
            length {int} -- Length of training loop, generally the number of batches in an epoch
                (i.e., the length of the dataloader).
        """
        self.mode = mode
        self.length = length

    def __call__(self, loss, metrics, i):
        track_str = '\r{:8s} | {:5d}/{:<5d}| '.format(self.mode, i + 1, self.length)
        loss_str = 'loss: {:9.4f} | '.format(loss)
        metric_str = ' | '.join('{}: {:9.4f}'.format(k, v) for k, v in metrics.items())
        print(track_str + loss_str + metric_str + '   ', end='')
        if i + 1 == self.length:
            print('')


def detach_objects(x):
    """Function to detach objects from gpu if x is a torch.Tensor or Iterable of torch.Tensors.
    Otherwise, returns the input.
  
    Arguments: 
        x {torch.Tensor or Iterable} -- Object to be detached, can be a torch tensor or an iterable
            of torch Tensors.
  
    Returns: 
        torch.Tensor or Iterable -- If input is a torch.Tensor, returns torch.Tensor. Or if input is
        iterable, returns Iterable. Otherwise, returns the input. 
    """
    if isinstance(x, torch.Tensor):
        return x.detach().cpu()
    elif isinstance(x, Iterable):
        out = []
        for x_i in x:
            if isinstance(x_i, torch.Tensor):
                out.append(x_i.detach().cpu())
            else:
                out.append(x_i)
        return out
    else:
        return x


class Runner(object):
    """Model trainer/evaluater.

    Switch between training and evaluation modes by setting the model.training attribute with
    model.train() and model.eval().

    Arguments:
        model {nn.Module} -- Model to train/evaluate.
        loss_fn {callable} -- Loss function with signature:
            fn(<model output>, <loader labels>) -> <torch scalar>

    Keyword Arguments:
        optimizer {torch.optim.Optimizer} -- Torch optimizer. Can be None if training will not
            be performed. (default: {None})
        scheduler {torch.optim.lr_scheduler._LRScheduler} -- Torch LR scheduler. Can be None if
            training will not be performed. (default: {None})
        batch_metrics {dict} -- Dict of (named) callables that calculate useful metrics. Each
            should have the same signature as the loss_fn. (default: {{'time': EPS()}})
        device {str or torch.device} -- Device for pytorch to use. (default: {'cpu'})
        writer {torch.utils.tensorboard.SummaryWriter} -- Tensorboard SummaryWriter.
            (default: {None})
        is_batch_scheduler {bool} -- Flag to call scheduler.step every batch instead of every epoch.
            (default: {False})
        logger {object} -- Arbitrary logger object with a call signature that matches the default
            Logger. If None, the default logger will be used. (default: {None})
    """

    def __init__(
        self, model, loss_fn=lambda *_: torch.tensor(0.), optimizer=None, scheduler=None,
        batch_metrics={'eps': EPS()}, device='cpu', writer=None, is_batch_scheduler=False, 
        logger=None
    ):
        self.model = model
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.batch_metrics = batch_metrics
        self.device = device
        self.writer = writer
        self.write_interval = 10
        self.epoch = 0
        self.iteration = 0
        self.history = {}
        self.latest = {}
        self.is_batch_scheduler = is_batch_scheduler
        self.logger = logger

    def __call__(self, loader, mode=None, return_preds=False):
        """Train or evaluate over an epoch of data.

        Arguments:
            loader {torch.utils.data.DataLoader} -- Torch data loader. The loader should return a
                tuple (x, y), where x can be passed directly to the model object (after loading to
                the correct device), and y can be passed directly to the loss and metric functions.

        Keyword Arguments:
            mode {str} -- Prefix for logging (text and tensorboard) (default: {None})
            return_preds {bool} -- Return targets and predictions for all samples in `loader`.
                (default: {False})

        Returns:
            None or tuple -- If `return_preds` is False, returns None. Otherwise, a tuple of the 
                model outputs and the targets.
        """

        assert len(loader) > 0, 'Loader is empty.'

        # Unpack
        model = self.model
        loss_fn = PooledMean(self.loss_fn)
        optimizer = self.optimizer
        scheduler = self.scheduler
        is_batch_scheduler = self.is_batch_scheduler
        batch_metrics = self.batch_metrics
        device = self.device

        # Reset metric cache's where required
        for k in batch_metrics.keys():
            if not isinstance(batch_metrics[k], Metric):
                batch_metrics[k] = PooledMean(batch_metrics[k])
            batch_metrics[k].reset()

        # Set logging prefix if not specified and get logger instance
        if mode is None:
            mode = 'train' if model.training else 'valid'
        if self.logger is None:
            logger = Logger(mode, length=len(loader))
        else:
            logger = self.logger

        if return_preds:
            y_pred_epoch = []
            y_epoch = []

        with torch.set_grad_enabled(model.training):

            for i_batch, (x, y) in enumerate(loader):
                if isinstance(x, torch.Tensor):
                    x = x.to(device)
                elif isinstance(x, Iterable):
                    x = [x_i.to(device) for x_i in x]
                else:
                    raise TypeError('First element returned by loader should be a tensor or list.')

                if isinstance(y, torch.Tensor):
                    y = y.to(device)
                elif isinstance(y, Iterable):
                    y = [y_i.to(device) for y_i in y]
                else:
                    raise TypeError('Second element returned by loader should be a tensor or list.')

                y_pred = model(x)
                loss_batch = loss_fn(y_pred, y)

                if model.training:
                    loss_batch.backward()
                    optimizer.step()
                    optimizer.zero_grad()
                    if scheduler is not None and is_batch_scheduler:
                        scheduler.step()
                    self.iteration += 1

                # Evaluate batch using metrics
                metrics_batch = {nm: fn(y_pred, y) for nm, fn in batch_metrics.items()}
                metrics_batch = {nm: v.detach().cpu() for nm, v in metrics_batch.items() if v is not None}
                metrics = {nm: fn.compute() for nm, fn in batch_metrics.items()}
                metrics = {nm: v.detach().cpu() for nm, v in metrics.items() if v is not None}
                loss = loss_fn.compute()

                if model.training and self.iteration % self.write_interval == 0:
                    self._write(loss_batch, metrics_batch, mode)

                # Log results
                logger(loss, metrics, i_batch)

                if return_preds:
                    y_pred_epoch.append(detach_objects(y_pred))
                    y_epoch.append(detach_objects(y))

        if model.training:
            if not is_batch_scheduler and scheduler is not None:
                scheduler.step()
            self.epoch += 1

        if not model.training:
            self._write(loss, metrics, mode)

        # Save loss and metric values in runner history attribute
        self.history[self.epoch] = self.history.get(self.epoch, {})
        self.history[self.epoch][mode] = {'loss': loss, 'metrics': metrics}
        self.latest = self.history[self.epoch][mode]

        # Combine batches (if feasible)
        if return_preds:
            if all(isinstance(y_i, torch.Tensor) for y_i in y_pred_epoch):
                y_pred_epoch = torch.cat(y_pred_epoch)
            if all(isinstance(y_i, torch.Tensor) for y_i in y_epoch):
                y_epoch = torch.cat(y_epoch)

            return y_pred_epoch, y_epoch

    def __str__(self):
        return (
            'Model training and evaluation runner\n\n'
            'Training and evaluation history:\n'
            '{\n' + ',\n'.join(f'  {k}:{v}' for k, v in self.history.items()) + '\n}'
        )

    def _write(self, loss, metrics, mode):
        if self.writer is None:
            return
        self.writer.add_scalar(f'loss/{mode}', loss.detach().cpu(), self.iteration)
        for metric_name, metric in metrics.items():
            self.writer.add_scalar(f'{metric_name}/{mode}', metric.detach().cpu(), self.iteration)

    def loss(self):
        return self.latest['loss']

    def metrics(self):
        return self.latest['metrics']
